/*  tty.c                                                       (CRC: c155398)
 *      terminal driver
 *
 *      relative path: ./drv
 *      creation date: 1992
 *      last modified: Thu Dec 19 18:18:17 1996
 *      header update: Fri Dec 20 09:09:11 1996
 *  
 *       Device drivers
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */


#include <sys/types.h>
#include <sys/resource.h>
#include <error.h>
#include <stdio.h>
#include <dir.h>
#include <event.h>


#define TTY_IBUFSIZE	256
#define NCC		13	/* number of control characters */

typedef struct {
	obj_t	bufin;
	obj_t	bufout;
	ushort	iflag;
	ushort	oflag;
	ushort	lflag;
	ushort	wid;
	ushort	hgt;
	obj_t	thread;
	uchar	cc[NCC];
	ushort	ipos;
	uchar	ibuf[TTY_IBUFSIZE];
} TTY;

/* bits for flags: */

#define TTY_ICRNL	0
#define TTY_INLCR	1
#define TTY_IGNCR	2
#define TTY_ICLOSED	14
#define TTY_ILNEXT	15

#define TTY_OCRNL	0
#define TTY_ONLCR	1
#define TTY_ONLRET	2
#define TTY_ONOCR	3
#define TTY_OSUSPEND	15

#define TTY_ECHO	0
#define TTY_ECHOE	1
#define TTY_ECHOK	2
#define TTY_ECHONL	3
#define TTY_ICANON	4
#define TTY_ISIG	5

/* special keys: */

#define VEOF	0
#define VEOL	1
#define VERASE	2
#define VINTR	3
#define VKILL	4
#define VQUIT	5
#define VSUSP	6
#define VSTART	7
#define VSTOP	8
#define VERASE2	9
#define VLNEXT	10
#define VWERASE 11
#define VRPRNT	12

extern label_t _icrnl, _inlcr, _igncr, _iflag;
extern label_t _ocrnl, _onlcr, _onocr, _onlret, _oflag;
extern label_t _echo, _echoe, _echok, _echonl, _icanon, _isig, _lflag;
extern label_t _thread, _wid, _hgt, _stdin, _stdout;
extern label_t _msleep, _signal, _suspend, _parent;

static obj_t tty_init(ObjFrame *frame, TTY *tty)
{
	frame->data = tty = (TTY*)_(OBJ_TASK, _mem_alloc, sizeof(TTY), ___, ___);
	frame->size = sizeof(TTY);
	tty->cc[VEOF] = ctrl('D'); 
	tty->cc[VEOL] = ctrl('J');
	tty->cc[VERASE] = ctrl('H');
	tty->cc[VERASE2] = 127;
	tty->cc[VINTR] = ctrl('C');
	tty->cc[VKILL] = ctrl('U');
	tty->cc[VQUIT] = ctrl('\\');
	tty->cc[VSUSP] = ctrl('Z');
	tty->cc[VSTART] = ctrl('Q');
	tty->cc[VSTOP] = ctrl('S');
	tty->cc[VLNEXT] = ctrl('V');
	tty->cc[VWERASE] = ctrl('W');
	tty->cc[VRPRNT] = ctrl('R');
	tty->wid = 80;
	tty->hgt = 24;
	tty->iflag = bit(TTY_ICRNL);
	tty->oflag = bit(TTY_ONLCR);
	tty->lflag = bit(TTY_ICANON)|bit(TTY_ECHO)|bit(TTY_ISIG);

	tty->thread = OBJ_NULL;
	tty->ipos = 0;
	return(OFLAG_KERNEL|OFLAG_HASVEC);
}

static obj_t tty_read(ObjFrame *frame, TTY *tty, uchar *buf, int len)
{
	if(tty->iflag & bit(TTY_ICLOSED)) {
		tty->iflag &= ~bit(TTY_ICLOSED);
		return(ERR_EOF);
	}
	return(err_(tty->bufin,_read,Ptr(buf),len,___));
}


static void tty_clearchar(TTY *tty, char ch)
{
	if(!(tty->lflag & bit(TTY_ECHO)))
		return;
	if(ch < 32)
		err_(tty->bufout,_write,Ptr("\b\b  \b\b"),6,___);
	else
		err_(tty->bufout,_write,Ptr("\b \b"),3,___);
}

static void tty_writechar(TTY *tty, char ch)
{
char cb[2];
	if(!(tty->lflag & bit(TTY_ECHO)))
		return;
	if(ch < 32 && ch != 9) {
		cb[0] = '^';
		cb[1] = ch+64;
		err_(tty->bufout,_write,Ptr(cb),2,___);
	} else 
		err_(tty->bufout,_write,Ptr(&ch),1,___);
}

static void tty_writebuf(TTY *tty)
{
int n = 0;
	while(n < tty->ipos) 
		tty_writechar(tty,tty->ibuf[n++]);
}

static void tty_clearline(TTY *tty)
{
int n = tty->ipos;
	while(n > 0) {
		n--;
		tty_clearchar(tty,tty->ibuf[n]);
	}
}


static obj_t tty_write(ObjFrame *frame, TTY *tty, uchar *buf, int len)
{
uchar cb[2048];
int i, count;
uchar ch;
	tty_clearline(tty);
	i = 0;
	count = 0;	
	while(len-- > 0) {
		ch = *buf++;
		if(ch == '\r' && (tty->oflag & bit(TTY_OCRNL))) {
			cb[i++] = '\n';
		} else if(ch == '\n' && (tty->oflag & bit(TTY_ONLCR))) {
			cb[i++] = '\r';
			cb[i++] = '\n';
		} else
			cb[i++] = ch;
		count++;
		while(tty->oflag & bit(TTY_OSUSPEND))
			_(OBJ_THREAD, _msleep, 0, ___, ___);
		if(!len || i > sizeof(cb)-2) {
			err_(tty->bufout,_write,Ptr(cb),i,___);
			i = 0;
		}
	}
	tty_writebuf(tty);
	return(count);
}

static obj_t tty_devread(ObjFrame *frame, TTY *tty, uchar *buf, int len)
{
	return(err_(tty->bufout,_read,Ptr(buf),len,___));
}

static obj_t tty_devwrite(ObjFrame *frame, TTY *tty, uchar *buf, int len)
{
int count;
uchar ch;
	count = 0;
	while(len-- > 0) {
		count++;
		ch = *buf++;
		if((tty->lflag & bit(TTY_ISIG)) && tty->thread != OBJ_NULL) {
			if(ch == tty->cc[VSUSP]) {
				obj_t parent = _(tty->thread, _get, _parent,
						  ___, ___);
				_(tty->thread, _suspend, ___, ___, ___);
				_(parent, _signal, EVENT_CHILDSTOP,
						 tty->thread, tty->thread);
				err_(tty->bufout,_write,Ptr("^Z\r\n"),4,___);
				continue;
			} 
			if(ch == tty->cc[VINTR]) {
				_(tty->thread, _free, ___, ___, ___);
				continue;
			}
		}
		if(!(tty->lflag & bit(TTY_ICANON))) {
			err_(tty->bufin,_write,Ptr(&ch),1,___);
			continue;
		}	
		if(tty->iflag & bit(TTY_ILNEXT)) {
			tty->iflag &= ~bit(TTY_ILNEXT);
			tty->ibuf[tty->ipos++] = ch;
			tty_writechar(tty, ch);
			continue;
		}
		if(ch == '\n' && (tty->iflag & bit(TTY_INLCR)))
			ch = '\r';
		if(ch == '\r' && (tty->iflag & bit(TTY_ICRNL)))
			ch = '\n';
		if(ch == tty->cc[VERASE] || ch == tty->cc[VERASE2] ||
		   ch == tty->cc[VWERASE] || ch == tty->cc[VKILL]) {	
			while(tty->ipos > 0) {
				tty->ipos--;
				tty_clearchar(tty, tty->ibuf[tty->ipos]);
				if(ch == tty->cc[VERASE] ||
				   ch == tty->cc[VERASE2] ||
				   (ch == tty->cc[VWERASE] &&
					(tty->ipos == 0 || 
					 tty->ibuf[tty->ipos-1] <= 32)))
					break;
			}
		} else if(ch == tty->cc[VRPRNT]) {
			_(tty->bufout, _write, Ptr("\r\n"), 2, ___);
			tty_writebuf(tty);
		} else if(ch == tty->cc[VEOL]) {
			tty->ibuf[tty->ipos++] = ch;
			_(tty->bufin, _write, Ptr(tty->ibuf), tty->ipos, ___);
			tty->ipos = 0;
			if(tty->lflag & (bit(TTY_ECHONL)|bit(TTY_ECHO)))
				_(tty->bufout, _write, Ptr("\r\n"), 2, ___);
		} else if(ch == tty->cc[VEOF]) {	
			tty->iflag |= bit(TTY_ICLOSED);
			_(tty->bufin, _write, Ptr("\n"), 1, ___);
			tty->ipos = 0;
		} else if(ch == tty->cc[VSTOP]) {
			tty->oflag |= bit(TTY_OSUSPEND);
		} else if(ch == tty->cc[VSTART]) {
			tty->oflag &= ~bit(TTY_OSUSPEND);
		} else if(ch == tty->cc[VLNEXT]) {
			tty->iflag |= bit(TTY_ILNEXT);
		} else {
			tty->ibuf[tty->ipos++] = ch;
			tty_writechar(tty, ch);
		}
	}
	return(count);
}

static MethodTable tty_method[] = {
	{&_init,(obj_method)tty_init},
	{&_read,(obj_method)tty_read},
	{&_write,(obj_method)tty_write},
	{&_devread,(obj_method)tty_devread},
	{&_devwrite,(obj_method)tty_devwrite}};

AttrTable tty_attr[] = {
	{&_stdin,offsetof(TTY,bufin),ATYPE_OBJ},
	{&_stdout,offsetof(TTY,bufout),ATYPE_OBJ},
	{&_icrnl,offsetof(TTY,iflag),ATYPE_FLAG|TTY_ICRNL},
	{&_inlcr,offsetof(TTY,iflag),ATYPE_FLAG|TTY_INLCR},
	{&_igncr,offsetof(TTY,iflag),ATYPE_FLAG|TTY_IGNCR},
	{&_iflag,offsetof(TTY,iflag),ATYPE_ULONG},
	{&_ocrnl,offsetof(TTY,oflag),ATYPE_FLAG|TTY_OCRNL},
	{&_onlcr,offsetof(TTY,oflag),ATYPE_FLAG|TTY_ONLCR},
	{&_onlret,offsetof(TTY,oflag),ATYPE_FLAG|TTY_ONLRET},
	{&_onocr,offsetof(TTY,oflag),ATYPE_FLAG|TTY_ONOCR},
	{&_oflag,offsetof(TTY,oflag),ATYPE_ULONG},
	{&_echo,offsetof(TTY,lflag),ATYPE_FLAG|TTY_ECHO},
	{&_echoe,offsetof(TTY,lflag),ATYPE_FLAG|TTY_ECHOE},
	{&_echok,offsetof(TTY,lflag),ATYPE_FLAG|TTY_ECHOK},
	{&_echonl,offsetof(TTY,lflag),ATYPE_FLAG|TTY_ECHONL},
	{&_icanon,offsetof(TTY,lflag),ATYPE_FLAG|TTY_ICANON},
	{&_isig,offsetof(TTY,lflag),ATYPE_FLAG|TTY_ISIG},
	{&_lflag,offsetof(TTY,lflag),ATYPE_ULONG},
	{&_wid,offsetof(TTY,wid),ATYPE_USHORT},
	{&_hgt,offsetof(TTY,hgt),ATYPE_USHORT},
	{&_thread,offsetof(TTY,thread),ATYPE_OBJ}};

static ObjClass Class_TTY = {
	tty_method,
	tty_attr,
	arraysize(tty_method),
	arraysize(tty_attr)};
 

obj_t Main(ObjFrame *frame, vaddr base, obj_t arg, obj_t env)
{
obj_t class_object;
obj_t class_tty;
	cprintf("Terminal Driver  v0.01  Nov 1992\n");
	class_object = dir_parse(OBJ_ROOT, "/class/Object");
	class_tty = _(class_object, _class, Ptr(&Class_TTY),___,___);
	dir_create(OBJ_ROOT,"/class/TTY", class_tty);
	return(OK);
}

