/*  vt100.c                                                     (CRC:1dcc7f11)
 *      standard vt100 emulation
 *      compatible with nearly any system
 *
 *      relative path: ./drv
 *      creation date: 1992
 *      last modified: Wed May 22 21:33:12 1996
 *      header update: Wed Dec 18 07:17:14 1996
 *  
 *       Device drivers
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */


#include <sys/vt100.h>

/* should be included from the code needing vt100 support,
   with the following macros defined before inclusion:
	VT_PUT_CHAR(vt,x,y,ch)
	VT_SCROLL_UP(vt,x1,y1,x2,y2,n)
	VT_SCROLL_DOWN(vt,x1,x2,y2,n)
	VT_CLEAR(vt,x1,y1,x2,y2)
	VT_SET_COLOR(vt)
	VT_SET_CURSOR(vt,x,y)
	VT_BEEP(vt) 
*/


static int vt_call_csi(struct vt_state *vt, char ch);
static int vt_call_ncsi(struct vt_state *vt, char ch);
static int vt_param(struct vt_state *vt,
		int n, int def, int min, int max);



static void vt_save_state(struct vt_state *vt)
{
	vt->save_xpos = vt->xpos;
	vt->save_ypos = vt->ypos;
	vt->save_y1 = vt->y1;
	vt->save_y2 = vt->y2;
	vt->save_attr = vt->attr;
	vt->save_fg = vt->fg;
	vt->save_bg = vt->bg;
}

static void vt_restore_state(struct vt_state *vt)
{
	vt->xpos = vt->save_xpos;
	vt->ypos = vt->save_ypos;
	vt->y1 = vt->save_y1;
	vt->y2 = vt->save_y2;
	vt->attr = vt->save_attr;
	vt->fg = vt->save_fg;
	vt->bg = vt->save_bg;
	VT_SET_COLOR(vt);
}



/* expects vt->wid, vt->hgt and implementation
 * variables to be initialized first: */

static void vt_init(struct vt_state *vt)
{
	vt->fg = 7;
	vt->bg = 0;
	vt->attr = 0;
	vt->mode = 0;
	vt->qmode = 0; 
	vt->y1 = 0;
	vt->y2 = vt->hgt-1;
	vt->xpos = 0;
	vt->ypos = 0;
	vt->esc = 0;
	vt->pcount = 0;
	vt_save_state(vt);
	VT_SET_COLOR(vt);
}


static void vt_write(register struct vt_state *vt, uchar *buf, int len)
{
uchar ch;
int x,n;
uchar *bp;
boolean check_scroll;
        while(len--) {
	    check_scroll = FALSE;	
            ch = *buf++;
	    switch(vt->esc) {
	    case 1:
		if(ch == '[') {
			vt->type = ch;
			vt->esc++;
			vt->value = 0;
			vt->pcount = 0;
			vt->ptype = 1;
			for(n=0; n<VT_MAXPARAM; n++)
				vt->param[n] = 0;
		} else {
			check_scroll = vt_call_ncsi(vt,ch);	
			vt->esc = 0;
		}	
		break;	
	    case 2:
		if(ch >= '0' && ch <= '9') 
			vt->value=(vt->value)*10+(ch-'0');
		else if(ch == '?')
			vt->ptype = -1;
		else {
			vt->param[vt->pcount++] = vt->value*vt->ptype;
			if(ch == ';') {
				if(vt->pcount >= VT_MAXPARAM)
					vt->pcount = VT_MAXPARAM-1;
				vt->value = 0;
			} else {
				check_scroll = vt_call_csi(vt,ch);
				vt->esc = 0;
			}
		}
		break;	
	    default:
		switch(ch) {
                case '\r':
                        vt->xpos = 0;
                        break;
                case '\n':
                        vt->ypos++;
			check_scroll = TRUE;
                        break;
                case '\b':
                        if (vt->xpos > 0)
                                vt->xpos--;
                        break;
                case '\t':
			n = (vt->xpos & ~7)+8;
			if(vt->mode & 0x10)
				VT_SCROLL_RIGHT(vt, vt->xpos,vt->ypos,
				  vt->wid-1,vt->ypos, n - vt->xpos);
                        vt->xpos = n;
			if(vt->xpos > vt->wid) {
				vt->xpos = 0; 
				vt->ypos++;
				check_scroll = TRUE;
			}
			break;
		case '\007':
			VT_BEEP(vt);
			break;
		case '\013':
			vt->xpos = 0;
			vt->ypos = vt->y1;
			break;
		case '\014':
			VT_CLEAR(vt,0,vt->y1,vt->wid-1,vt->y2);
			break;
		case '\033':
			vt->esc++;
			break;
		case '\205':
			vt->xpos = 0;
		case '\204':
			vt->ypos++;
			check_scroll = TRUE;
			break;
		case '\210':	/* XXX - set a tabstop */
			break;	
		case '\215':
			vt->ypos--;
			check_scroll = TRUE;
			break;
		case '\216':	/* XXX - map G2 into GL for next char only */
		case '\217':	/* XXX - map G3 into GL for next char */
		case '\220':	/* XXX - device control string */
		case '\221':	/* XXX - start of string - ignored */
		case '\222':	/* XXX - device attribute request */
			break;
		case '\223':
			vt->esc = 2;
			vt->type = '[';
			vt->esc++;
			vt->value = 0;
			vt->pcount = 0;
			vt->ptype = 1;
			for(n=0; n<VT_MAXPARAM; n++)
				vt->param[n] = 0;
                default:
			if(vt->mode & 0x10)
				VT_SCROLL_RIGHT(vt, vt->xpos,vt->ypos,
				  vt->wid-1,vt->ypos,1);	
			if(ch<32 || ch >126)
				break;
			VT_PUTCHAR(vt,vt->xpos,vt->ypos,ch);
                        if (++(vt->xpos) >= vt->wid) {
				if(vt->qmode & 0x80)
					vt->xpos = vt->wid-1; 
				else {	
                                	vt->xpos = 0;
                                	vt->ypos++;
					check_scroll = TRUE;
				}
                        }
                }
	    }
	    if(check_scroll) {
		if (vt->ypos == vt->y2+1 || vt->ypos >= vt->hgt) {
			x = len;
			n = 1;
			bp = buf;
			while(x-- > 0 && n < (vt->y2-vt->y1) && 
			 *bp != '\033') 
				if (*bp++ == '\n')
					n++;
               		vt->ypos = vt->y2-n+1;
			VT_SCROLL_UP(vt,0,vt->y1,vt->wid-1,vt->y2,n);
        	} else if (vt->ypos == vt->y1-1) {
			vt->ypos = vt->y1;
			VT_SCROLL_DOWN(vt,0,vt->y1,vt->wid-1,vt->y2,1);
		} else if (vt->ypos < 0)
			vt->ypos = 0;
	    }
	    if(vt->xpos < 0)
		vt->xpos = 0;
	    else if(vt->xpos >= vt->wid)
		vt->xpos = vt->wid-1;
	    if(vt->ypos < 0)
		vt->ypos = 0;
	    else if(vt->ypos >= vt->hgt)
		vt->ypos = vt->hgt-1;
	}
	VT_SET_CURSOR(vt, vt->xpos, vt->ypos);
}




static int vt_call_csi(struct vt_state *vt, char ch)
{
register int i, n;
switch(ch) {
case 'A':
	vt->ypos -= vt_param(vt,1,1,1,vt->hgt);
	break;
case 'B':
	vt->ypos += vt_param(vt,1,1,1,vt->hgt);
	break;
case 'C':
	vt->xpos += vt_param(vt,1,1,1,vt->wid);
	break;
case 'D':
	vt->xpos -= vt_param(vt,1,1,1,vt->wid);
	break;
case 'f':
case 'H':
	vt->ypos = vt_param(vt,0,1,1,vt->hgt)-1;
	vt->xpos = vt_param(vt,1,1,1,vt->wid)-1;
	break;
case 'J':
	switch(vt->param[0]) {
	case 0: VT_CLEAR(vt,vt->xpos,vt->ypos,vt->wid-1,vt->ypos);
		VT_CLEAR(vt,0,vt->ypos+1,vt->wid-1,vt->y2); break;
	case 1: VT_CLEAR(vt,0,0,vt->wid-1,vt->ypos-1); 
	        VT_CLEAR(vt,0,vt->ypos,vt->xpos,vt->ypos); break;
	case 2: VT_CLEAR(vt,0,vt->y1,vt->wid-1,vt->y2);	break;
	}
	break;
case 'K':
	switch(vt->param[0]) {
	case 0: VT_CLEAR(vt,vt->xpos,vt->ypos,vt->wid-1,vt->ypos); break;
	case 1: VT_CLEAR(vt,0,vt->ypos,vt->xpos,vt->ypos); break;
	case 2: VT_CLEAR(vt,0,vt->ypos,vt->wid-1,vt->ypos); break;
	}
	break;
case 'L':
	n = vt_param(vt,0,1,1,vt->hgt);
	VT_SCROLL_DOWN(vt,0,vt->ypos,vt->wid-1,vt->y2,n);	
	break;
case 'M':
	n = vt_param(vt,0,1,1,vt->hgt);
	VT_SCROLL_UP(vt,0,vt->ypos,vt->wid-1,vt->y2,n);	
	break;
case '@':
	n = vt_param(vt,0,1,1,vt->wid-1-vt->xpos);
	VT_SCROLL_RIGHT(vt,vt->xpos,vt->ypos,vt->wid-1,vt->ypos,n);	
	break;
case 'P':
	n = vt_param(vt,0,1,1,vt->wid-1-vt->xpos);
	VT_SCROLL_LEFT(vt,vt->xpos,vt->ypos,vt->wid-1,vt->ypos,n);	
	break;
case 'X':
	n = vt_param(vt,0,1,1,vt->wid-1-vt->xpos);
	VT_CLEAR(vt,vt->xpos,vt->ypos,vt->xpos+n-1,vt->ypos);
	break;
case 'm':
	if(vt->pcount == 0)
		vt->pcount++;
	for(i=0; i<vt->pcount; i++) {
		n = vt->param[i];
		if(!n) {
			vt->attr = 0; 
			vt->fg = 7;
			vt->bg = 0;
		} else if (n < 16)
			vt->attr |= (1<<n);
		else if (n < 28)
			vt->attr &= ~(1<<(n-20));
		else if (n < 38)
			vt->fg = n-30;
		else if (n < 48)
			vt->bg = n-40;
		else if (n < 58)
			vt->fg = n-50+8;
		else if (n < 68)
			vt->bg = n-60+8;
	}
	VT_SET_COLOR(vt);
	break;
case 'c':
	if(vt->wid >= 132)
		printf("\033[?61;1;6c");
	else
		printf("\033[?61;6c");
	break;
case 'n': 
	n = vt_param(vt,0,0,0,9);
	switch(n) {
	case 5:
		printf("\033[0n");
	case 6:
		printf("\033[%d;%dR",vt->ypos+1,vt->xpos+1);
		break;
	default:
		break;
	}
	break; 
case 'r':
	vt->y1 = vt_param(vt,0,1,1,vt->hgt)-1;
	vt->y2 = vt_param(vt,1,vt->hgt,1,vt->hgt)-1;
	break;
case 's':
	vt_save_state(vt);
	break;
case 'u':
	vt_restore_state(vt);
	break;
case 'h':
	for(i=0; i<vt->pcount; i++) {
		n = vt->param[i];
		if(n >= 0)
			vt->mode |= (1<<n);
		else
			vt->qmode |= (1<<(-n));
	}
	break;
case 'l':
	for(i=0; i<vt->pcount; i++) {
		n = vt->param[i];
		if(n >= 0)
			vt->mode &= ~(1<<n);
		else
			vt->qmode &= ~(1<<(-n));
	}
	break;
 }
return(FALSE);
}

static int vt_call_ncsi(struct vt_state *vt, char ch)
{
switch(ch) {
case 'E':
	vt->xpos = 0;
case '9':
case 'D':
	vt->ypos++;
	return(TRUE);
case 'H':	/* XXX - horizontal tab set */
	break;
case '6':
case 'M':
	vt->ypos--;
	return(TRUE);
case '7':
	vt_save_state(vt);
	break;
case '8':
	vt_restore_state(vt);
	break;
case '=':
	break;
case '>':
	break;
case '#':
	break;
case '(':
	break;
case ')':
	break;
 }
return(FALSE);
}


static int vt_param(struct vt_state *vt,
		int n, int def, int min, int max)
{
register int param;
	param = vt->param[n];
	if(param == 0)
		param = def;
	if(param < min)
		param = min;
	if(param > max)
		param = max;
	return(param);
}

