/*  object.h                                                    (CRC:dfee41c1)
 *      relative path: ./include
 *      creation date: 1992
 *      last modified: Wed May 22 21:35:30 1996
 *      header update: Wed Dec 18 07:17:26 1996
 *  
 *       Include files
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#ifndef __OBJECT_H__
#define __OBJECT_H__

#ifndef ASM

#include <sys/types.h>

/* calling frame: */
typedef struct obj_frame {
	obj_t	self;
	obj_t	op;	
	obj_t	oid;
	obj_t	next;
	void	*data;
	ulong	size;
} ObjFrame;



typedef obj_t (*obj_method)(ObjFrame *frame, void *data, obj_t arg1, obj_t arg2, obj_t arg3);


/* attribute table for quick lookups: */
typedef struct attr_table {
	label_t	*id;		/* pointer to label id */
	ushort	value;		/* offset into structure */
	uchar	type;	
} AttrTable;	


/* method table: */
typedef struct method_table {
	label_t	   *id;		/* pointer to label id */
	obj_method func;	/* function pointer */
} MethodTable; 


/* class definition: */
typedef struct obj_class {
	MethodTable	*method;
	AttrTable	*attr;
	ushort		nmethod;
	ushort		nattr;
} ObjClass;


#ifdef KERNEL
typedef struct object {
	unsigned flags:8;	/* object flags */
	unsigned ref:24;	/* reference count */
	ObjClass *class;	/* class definition */
	void	*data;		/* data pointer/value */
	ulong	size;		/* size of data */
	obj_t	next;		/* next object in chain */
} Object;
#endif


#define ATYPE_TYPE	0x3f	/* type mask */
#define ATYPE_CHAR	0x01
#define ATYPE_UCHAR	0x02
#define ATYPE_SHORT	0x03
#define ATYPE_USHORT	0x04
#define ATYPE_LONG	0x05
#define ATYPE_ULONG	0x06
#define ATYPE_OBJ	0x07
#define ATYPE_FLAG	0x20
#define ATYPE_NOREAD	0x40
#define ATYPE_NOWRITE	0x80

#endif

/* object flags: */
#define OFLAG_NONE	0x00	/* no object flags */
#define OFLAG_HASVEC	0x01	/* object has a vector */
#define OFLAG_VECTOR	0x02	/* object is simple vector type */
#define OFLAG_FORWARDER	0x04	/* object is a forwarder */
#define OFLAG_NOREAD	0x20	/* no read access */
#define OFLAG_NOWRITE 	0x40	/* no write access */
#define OFLAG_KERNEL	0x80	/* kernel level object */

/* the object descriptor (type 'obj_t') is defined as follows:
 *  bits 31-28:  type 		(4 bits: range 0-15)
 *  bits 27-0:   data/index	(28 bits: range 256meg)
 */

/* the following types are imposed by the kernel
 */
#define TYPE_INT	0x0	/* positive integer (or pointer) */
#define TYPE_CHAR	0x1	/* (up to 16 bit) character */
#define TYPE_SSTR	0x2	/* short string (4 7-bit characters) */

#define TYPE_OBJ	0x8	/* object: data is object index (per task) */
#define TYPE_TPTR	0x9	/* pointer in task space */
#define TYPE_SPTR	0xa	/* pointer in stack space */
#define TYPE_KPTR	0xb	/* pointer in kernel space */
#define TYPE_LABEL	0xd	/* system-wide label or reference */
#define TYPE_ERR	0xe	/* error code */
#define TYPE_NINT	0xf	/* negative integer (or kernel pointer) */

#define MAX_DATA	0x0fffffff	/* maximum data value */


#define	Obj(t,d)	(obj_t)(((ulong)(t) << 28) | ((ulong)(d) & MAX_DATA))
#define objtype(oid)	((ulong)(oid)>>28)
#define objdata(oid)	((ulong)(oid)&MAX_DATA)
#define ObjErr(d)	Obj(TYPE_ERR,d)
#define err(oid)	((obj_t)(oid) >= ObjErr(0) \
			&& (obj_t)(oid) <= ObjErr(MAX_DATA))
#define Char(c)		Obj(TYPE_CHAR,c);
#define Int(d)		(((d)<0)?Obj(TYPE_NINT,(d)):Obj(TYPE_INT,(d)))
#define Ptr(p)		(obj_t)(p)
#define Str(s)		Obj(TYPE_SSTR,((s)[0])|((s)[1]<<7)	\
				|((s)[2]<<14)|((s)[3]<<21)))

/* built-in special objects: */

#define OBJ_NULL	Obj(TYPE_LABEL,0x00)		
#define OBJ_TRUE	Obj(TYPE_LABEL,0x01)
#define OBJ_FALSE	Obj(TYPE_LABEL,0x02)
#define OBJ_ANY		Obj(TYPE_LABEL,0x03)
#define OBJ_LABEL_MIN	Obj(TYPE_LABEL,0x100)

#define OBJ_SYS		Obj(TYPE_OBJ, MAX_DATA)		/* system calls */
#define OBJ_KERNEL	Obj(TYPE_OBJ, MAX_DATA-1)	/* kernel task */
#define OBJ_TASK	Obj(TYPE_OBJ, MAX_DATA-2)	/* current task */
#define OBJ_THREAD	Obj(TYPE_OBJ, MAX_DATA-3)	/* current thread */
#define OBJ_ROOT	Obj(TYPE_OBJ, MAX_DATA-4)	/* root directory object */
#define OBJ_CWD		Obj(TYPE_OBJ, MAX_DATA-5)	/* current working directory */
#define OBJ_THREAD_MIN	(MAX_DATA-0x100)
#define OBJ_THREAD_MAX	(MAX_DATA-0x100+3)
#define OBJ_STDIN	Obj(TYPE_OBJ, OBJ_THREAD_MIN+0)	/* standard input */
#define OBJ_STDOUT	Obj(TYPE_OBJ, OBJ_THREAD_MIN+1)	/* standard output */
#define OBJ_STDERR	Obj(TYPE_OBJ, OBJ_THREAD_MIN+2)	/* standard error */


#ifndef ASM

extern label_t _init, _new, _free, _eval, _dup, _type;
extern label_t _class, _next, _prev;
extern label_t _get, _set, _getattr, _create, _remove;
extern label_t _read, _write, _readat, _writeat;
extern label_t _devread, _devwrite;
extern label_t _mem_alloc, _mem_free;
extern label_t _vm_alloc, _vm_free, _vm_map, _vm_reserve, _vm_share;

extern obj_t ObjCall(obj_t oid, obj_t op, obj_t arg1, obj_t arg2, obj_t arg3);
extern obj_t ObjCall_Trap(obj_t oid, obj_t op, obj_t arg1, obj_t arg2, obj_t arg3);
extern obj_t ObjCall_Self(obj_t oid, obj_t op, obj_t arg1,obj_t arg2,obj_t arg3, obj_t self);
extern ulong AttrLookup(AttrTable *table, obj_t id);
extern obj_t AttrGet(void *data, AttrTable *table, int tsize, obj_t id);
extern obj_t AttrSet(void *data, AttrTable *table, int tsize, obj_t id, obj_t val);
extern obj_t Label(char *str);
extern obj_t LabelRegister(char *str, obj_t value, obj_t class);
extern obj_t LabelClassify(obj_t id, obj_t class);
extern obj_t ObjString(char *buf, int len, obj_t id);
extern void obj_print(obj_t oid);

extern obj_t Main(ObjFrame *frame, vaddr base, obj_t arg, obj_t env);

#endif


/* shortened form of object_call: */
#define _ ObjCall_Trap
#define err_ ObjCall 
#define ___ OBJ_NULL

#endif /* __OBJECT_H__ */

