/*  buffer.c                                                    (CRC:195727ad)
 *      relative path: ./sys
 *      creation date: 1992
 *      last modified: Wed Dec 18 07:18:03 1996
 *      header update: Wed Dec 18 07:18:04 1996
 *  
 *       Microkernel system files
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* buffer.c
 *  generic buffering routines
 */

#define KERNEL
#include <sys/types.h>
#include <sys/resource.h>
#include <event.h>
#include <object.h>
#include <sys/global.h>
#include <error.h>
#include <dir.h>


typedef struct buffer {
	struct buffer	*next;		/* next buffer */
	ZoneP		zone;		/* zone to allocate from */
	ushort		ctrl;		/* blocking control flags */
	ushort		nodesize;	/* number of characters per node */
	ulong		nchars;		/* number of characters in buffer */
	ulong		maxsize;	/* maximum size before blocking */
	uchar		*head;		/* pointer to head */
	uchar		*tail;		/* pointer to tail */
	ulong		param;		/* usage-based parameter */
	obj_t		rwait_thread;	/* thread waiting to read */
	obj_t		wwait_thread;	/* thread waiting to write */
	lock_t		rlock;		/* buffer read lock */
	lock_t		wlock;		/* buffer write lock */
} Buffer;


extern label_t _param, _ctrl, _maxsize, _size; 
extern label_t _rnoblock, _wnoblock;


#define BCTRL_RNOBLOCK	0
#define BCTRL_WNOBLOCK	1


static ZoneP buffer_zone;
static ZoneP charbuf_zone;


static obj_t buffer_init(ObjFrame *frame, Buffer *buf)
{
	frame->data = buf = (Buffer*)zalloc(buffer_zone);
	frame->size = sizeof(Buffer);
	buf->zone = charbuf_zone;
	buf->nodesize = CHAR_BB_SIZE;
	buf->nchars = 0;
	buf->tail = NULL;
	buf->head = NULL;
	buf->ctrl = 0;		/* block for everything */
	buf->rwait_thread = 0;
	buf->wwait_thread = 0;
	buf->maxsize = MAXBUFSIZE;
	LOCK_CLEAR(buf->rlock);
	LOCK_CLEAR(buf->wlock);
	return(OFLAG_HASVEC|OFLAG_KERNEL);
}


static obj_t buffer_free(ObjFrame *frame, Buffer *buf)
{
vaddr node, oldnode;
	LOCK(buf->wlock)
	LOCK(buf->rlock)
	node = ((vaddr)(buf->head) & ~(buf->nodesize-1));
	while(node) {
		oldnode = node;
		node = *(vaddr*)(node+buf->nodesize-sizeof(vaddr));	
		zfree(buf->zone, (void*)oldnode);
	}
	zfree(buffer_zone, buf);
	/* no need to free locks, since buffer is now gone */
	return(OK);
}




static obj_t buffer_read(ObjFrame *frame, Buffer *buf, vaddr adr, int count)
/* XXX - Needs semaphores
 */
{
register long n;
register uchar *head;
uchar **link;
Event *event;
int cflag;
	if(!count)
		return(0);
	LOCK(buf->rlock)
	while(!buf->nchars && int_level == 0
	 && !(buf->ctrl & bit(BCTRL_RNOBLOCK))) {
		buf->rwait_thread = thread->self;
		waitfor(EVENT_BUFIN, PARAM_ANY);
		buf->rwait_thread = 0;
	}
	head = buf->head;
	if (count > buf->nchars)
		count = buf->nchars;
	n = 0;
	link = (uchar**)(align((vaddr)head+1,buf->nodesize)-sizeof(uchar*));
	while(count--) {
		if(!head || head >= (uchar*)link) {
			if (head && *link) {
				head = *link;
				zfree(buf->zone, (void*)((vaddr)link
					 	& ~(buf->nodesize-1)));
				link = (uchar**)(align((vaddr)head+1,
					buf->nodesize)-sizeof(uchar*));
			} else
				panic(0, "Scrambled buffers");
		}
		*(char*)(adr++) = *head++;
		n++;
	}
	CRITICAL_BEGIN(cflag)
	if(!(buf->nchars -= n) && n && GETLOCK(buf->wlock)<0) {
		if (buf->tail == head)
			buf->tail = head = 
				(uchar*)((vaddr)head & ~(buf->nodesize-1));
		UNLOCK(buf->wlock);
	}
	CRITICAL_END(cflag)
	buf->head = head;
	if(buf->wwait_thread) {
		event = (Event*)zalloc(event_zone);
		event->event = EVENT_BUFOUT;
		event->param = PARAM_ANY;
		event->sender = thread->self;
		event->receiver = buf->wwait_thread;
		event_send(event);
	}
	UNLOCK(buf->rlock);
	return(n);
}


static obj_t buffer_write(ObjFrame *frame, Buffer *buf, vaddr adr, int count)
{
register int n;
register uchar *tail;
uchar **link;
Event *event;
int nchars;
int cflag;
	if(!count)
		return(0);
	LOCK(buf->wlock)
	while(buf->nchars > buf->maxsize && int_level == 0
	 && !(buf->ctrl & bit(BCTRL_WNOBLOCK))) {
		buf->wwait_thread = thread->self;
		thread->max_quantum = kinfo.minquantum;
		waitfor(EVENT_BUFOUT, PARAM_ANY);
		buf->wwait_thread = 0;
	}	
	tail = buf->tail;
	n = 0;
	nchars = 0;
	if(tail)
		link = (uchar**)(align((vaddr)tail+1,buf->nodesize)
				-sizeof(uchar*));
	else
		link = &buf->head;
	while(count--) {
		if(!tail || tail >= (uchar*)link) {
			if (!(*link = (uchar*)zalloc(buf->zone))) {
				panic(0, "buf alloc failed");
				break;	/* no more buffers available */
			}
			tail = *link;
			link = (uchar**)(align((vaddr)tail+1,buf->nodesize)
				-sizeof(uchar*));
		}
		*tail++ = *(char*)(adr++);
		nchars++;
		n++;
	}
	CRITICAL_BEGIN(cflag)
	buf->tail = tail;
	buf->nchars += nchars;
	CRITICAL_END(cflag)
	if(buf->rwait_thread) {
		event = (Event*)zalloc(event_zone);	
		event->event = EVENT_BUFIN;
		event->param = PARAM_ANY;
		event->sender = thread->self;
		event->receiver = buf->rwait_thread;
		event_send(event);
	}
	UNLOCK(buf->wlock)
	return(n);
}


static MethodTable buffer_method[] = {
	{&_init,(obj_method)buffer_init},
	{&_free,(obj_method)buffer_free},
	{&_read,(obj_method)buffer_read},
	{&_write,(obj_method)buffer_write}};

static AttrTable buffer_attr[] = {
	{&_rnoblock,offsetof(Buffer,ctrl),ATYPE_FLAG|BCTRL_RNOBLOCK},
	{&_wnoblock,offsetof(Buffer,ctrl),ATYPE_FLAG|BCTRL_WNOBLOCK},
	{&_ctrl,offsetof(Buffer,ctrl),ATYPE_USHORT},
	{&_param,offsetof(Buffer,param),ATYPE_ULONG},
	{&_maxsize,offsetof(Buffer,maxsize),ATYPE_ULONG},
	{&_size,offsetof(Buffer,nchars),ATYPE_ULONG}};

static ObjClass Class_Buffer = {
	buffer_method,
	buffer_attr,
	arraysize(buffer_method),
	arraysize(buffer_attr)};


void Buffer_init(void)
{
	buffer_zone = (ZoneP)zinit(sizeof(Buffer),"Buffer");
	charbuf_zone = (ZoneP)zinit(CHAR_BB_SIZE, "charbuf");
	class_buffer = _(class_object, _class, Ptr(&Class_Buffer),___,___);
	dir_create(kinfo.Root,"/class/Buffer", class_buffer);
}

