/*  event.c                                                     (CRC:58c67ffc)
 *      relative path: ./sys
 *      creation date: 1992
 *      last modified: Wed Dec 18 07:18:07 1996
 *      header update: Wed Dec 18 07:18:07 1996
 *  
 *       Microkernel system files
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* event.c
 *  event sending/receiving
 */

#define KERNEL
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/global.h>
#include <event.h>
#include <error.h>



status_t event_send(Event *event)
/* before invoking, event->sender and event->receiver should
 * be set up, as well as event->event and event->param
 */
{
Thread *node;
boolean found;
boolean b;
	CSWITCH_OFF(b)
	node = (Thread*)runq.prev;	
	found = FALSE;
	/* reverse search to catch current thread first */
	while((Queue*)node != &runq) {
		if(node->self == event->receiver)
			if(event->event == EVENT_TRAP) {
				kprintf("XXX: no trap handler\n");
				CSWITCH_ON(b)
				return(OK);
			} else {
				found = TRUE;
				break;
			}
		node = node->prev;
	}
/* XXX - potential problem -- if
   the thread receives other events after the wakeup, then
   wakes up, the first event read won't be the wakeup
   event -- will this cause problems?  
 */
	if(!found) {
		node = (Thread*)waitq.next;
		while((Queue*)node != &waitq) {
			if(node->self == event->receiver) {
				found = TRUE;
				break;
			}
			node = node->next;
		}
	}
	if (found) {
		qaddtail(&node->eventq, (Queue*)event);	
		if(((event->event & node->w_event)
		  && (event->param == node->w_param
		  || node->w_param == PARAM_ANY)) 
		  | (event->event == EVENT_WAKEUP)) {
			qremnode((Queue*)node);	
			kinfo.tcount++;
			node->w_event = 0;
			node->w_param = 0;
			qaddhead(&runq,(Queue*)node);
		} 
		if(event->event == EVENT_TERMINATE) {
			qremnode((Queue*)node);
			if(!node->w_event)
				kinfo.tcount--;
			node->w_param = 0;
			qaddtail(&deadq,(Queue*)node);
			qremnode((Queue*)cleanup_thread);
			qaddhead(&runq, (Queue*)cleanup_thread);
		}
		CSWITCH_ON(b)
		return(OK);
	} else {
		CSWITCH_ON(b)
		return(ERR_NOTHREAD);
	}
}


Event *event_recv(ulong evmask, ulong param)
{
Event *event;
boolean b;
	CSWITCH_OFF(b)
	while(1) {
		/* see if any events were waiting: */
		while((Queue*)(event = (Event*)qremhead(&thread->eventq))
	   			!= &thread->eventq) {
			if((event->event & evmask) && 
			 ((event)->param == param || param == PARAM_ANY)) {
				CSWITCH_ON(b)
				return(event);
			}
			zfree(event_zone, event);	
		}
		thread->max_quantum += kinfo.quantum_wait;
		if(thread->max_quantum > kinfo.maxquantum)
			thread->max_quantum = kinfo.maxquantum;
		thread->w_event = evmask;
		thread->w_param = param;
		/* unlink us from the run-queue */
		qremnode((Queue*)thread);
		kinfo.tcount--;
		/* link us in the wait queue */
		qaddhead(&waitq,(Queue*)thread);
		while(qempty(&thread->eventq)) 
			thread_next();
		/* we're alive and we've got an event.. */	
	}
}


ulong waitfor(ulong evmask, ulong param)
{
Event *event;
	event = event_recv(evmask,param);	
	param = event->param;
	zfree(event_zone, event);
	return(param);
}

/* lock format:
 *  bit 0: 1=lock is locked
 *  bit 1: 1=someone is waiting for the lock
 *  bits 2-7: count of how many waiting (up to 64)
 */

void lock_wait(volatile lock_t *lock)
{
int cflag;
	CRITICAL_BEGIN(cflag)
	if((*lock)&1) {
		if(int_level > 0)
			panic(0,"Lock(%x) is locked in interrupt",lock);
		if(!thread)
			panic(0,"Uninitialized lock(%x)",lock); 
		(*lock) = ((*lock)+4) | 2;
		thread->w_event = EVENT_LOCK;
		thread->w_param = (ulong)lock;
		kinfo.tcount--;
		qremnode((Queue*)thread);
		qaddhead(&waitq,(Queue*)thread);
		while((*lock)&1) {
			thread_next();
			CRITICAL_END(cflag)
			CRITICAL_BEGIN(cflag)
			if(((*lock) -= 4) < 4)
				(*lock) &= ~2;
			CRITICAL_END(cflag)
		}
	} else {
		(*lock) |= 1;
		CRITICAL_END(cflag)
	}
}

int lock_get(volatile lock_t *lock)
{
int cflag;
int n;
	CRITICAL_BEGIN(cflag)
	if((*lock)&1) 
		n = (*lock)>>2;
	else {
		(*lock) |= 1;
		n = -1;
	}
	CRITICAL_END(cflag)
	return(n);
}


void lock_signal(volatile lock_t *lock)
{
int cflag;
boolean b;
Thread *node,*next;
	CSWITCH_OFF(b)
	CRITICAL_BEGIN(cflag)
	(*lock) &= ~1;
	CRITICAL_END(cflag)
	if((*lock)&2) {
		node = (Thread*)waitq.next;
		while((Queue*)node != &waitq) {
			next = node->next;
			if(node->w_event == EVENT_LOCK 
			&& node->w_param == (ulong)lock) {
				qremnode((Queue*)node);	
				node->w_event = 0;
				node->w_param = 0;
				kinfo.tcount++;
				qaddhead(&runq,(Queue*)node);
			}	
			node = next;
		}
	}
	CSWITCH_ON(b)
}

void event_init(void)
{
	event_zone = (ZoneP)zinit(sizeof(Event),"event");
	LOCK_CLEAR(eventlock);
}
