/*  $Id: kmem.c,v 1.1 1996/11/12 01:44:31 shaggy Exp shaggy $
 *      kernel memory allocation/management
 *      relative path: ./sys
 *  
 *       Microkernel system files
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/*  $Log: kmem.c,v $
 *  Revision 1.1  1996/11/12  01:44:31  shaggy
 *  Initial revision
 *
 *      creation date: 1992
 */


#define KERNEL
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/global.h>
#include <error.h>
#include <event.h>
#include <stdio.h>


extern ulong etext(void), edata(void), end(void), sfs_base;

static vaddr memlo;
static vaddr memhi;


/* kernel memory allocation routines:
 */

void *task_mem_malloc(Task *task, int size)
{
MemList **ml;
MemList *node,*newnode;
void *adr;
ulong accmode;
	size = align(size,sizeof(MemList))+sizeof(MemList);
	LOCK(task->memlock)
	ml = &task->memfree;
	while(*ml && ((*ml)->size < size)) 
		ml = &((*ml)->free);
	node = *ml;
	if(!node) {
		if(task == &kernel)
			accmode = VM_KRW;
		else
			accmode = VM_URW;
		/* XXX - how should the access mode be determined? */
		*ml = node = (MemList*)vm_alloc(&task->data,
				align(size,PAGE_SIZE), accmode);
		node->size = align(size,PAGE_SIZE);
		node->free = NULL;
		qaddtail(&task->mem, (Queue*)node);
	}
	if(node->size > size)  {
		newnode = (MemList*)((vaddr)node + size);
		newnode->size = node->size-size;
		newnode->free = node->free;
		qaddhead((Queue*)node,(Queue*)newnode);
		node->free = newnode;
		node->size = size;
	} 
	*ml = node->free;
	node->free = NULL;
	UNLOCK(task->memlock)
	adr = (void*)((vaddr)node+sizeof(MemList));
	return(adr);	
}

void task_mem_free(Task *task, void *adr)
{
MemList *node;
	LOCK(task->memlock)
	node = (MemList*)((vaddr)adr-sizeof(MemList));
	if(node->free || node->size < sizeof(MemList))
		panic(0,"invalid free(%x)",adr);
	node->free = task->memfree;
	task->memfree = node;
	UNLOCK(task->memlock)
}


void *kmalloc(int size)
{
	return(task_mem_malloc(&kernel,size));
}


void kfree(void *adr)
{
	task_mem_free(&kernel,adr);
}	



void ksfree(void *adr, int size)
{
MemList *node;
vaddr va;
	LOCK(kernel.memlock)
	va = align((vaddr)adr,sizeof(MemList));
	size -= (va-(vaddr)adr);
	kinfo.wasted_bytes += (va-(vaddr)adr);
	node = (MemList*)adr;
	node->size = size & ~(sizeof(MemList)-1);
	kinfo.wasted_bytes += (size-node->size);
	if(node->size < sizeof(MemList)) {
		kinfo.wasted_bytes += node->size;
		UNLOCK(kernel.memlock)
		return;
	}
	node->free = kernel.memfree;
	kernel.memfree = node;
	qaddtail(&kernel.mem, (Queue*)node);
	UNLOCK(kernel.memlock)
}


void *ksmalloc(int size)
{
vaddr adr;
	memlo = align(memlo, 4);
	if(memlo+size > memhi) {
		kinfo.wasted_bytes += (memhi-memlo);
		memlo = vm_alloc(&kernel.data, size, VM_KRW);
		memhi = memlo + align(size,PAGE_SIZE);
	}	
	adr = memlo;
	memlo += size;
	return((void*)adr);
}



void mem_status(void)
{
int totsize;
int freesize;
int freenodes;
int totnodes;
MemList *node;
	node = (MemList*)kernel.mem.next;
	totsize = 0;
	freesize = 0;
	totnodes = 0;
	freenodes = 0;
	while(node != (MemList*)&kernel.mem) {
		totnodes++;
		totsize += node->size;
		node = node->next;
	}
	node = kernel.memfree;
	while(node) {
		freenodes++;
		freesize += (node->size-sizeof(MemList));
		node = node->free;
	}
	kprintf("avail: %d out of %d (%d/%d)\n", freesize, totsize, freenodes, totnodes);
}

void mem_init(void)
{
	kinfo.wasted_bytes = 0;
	qclear(&kernel.mem);
	kernel.memfree = NULL;
	LOCK_CLEAR(kernel.memlock);
	memlo = 0;
	memhi = 0;
}


void mem_list(void)
{
MemList *node;
	node = (MemList*)kernel.mem.next;
	while(node != (MemList*)&kernel.mem) {
		printf("%8x: size:%8x prev:%8x next:%8x free:%8x\n",
			(uint)node, (uint)node->size, (uint)node->prev,
			(uint)node->next, (uint)node->free);
		node = node->next;
	}
}

