/*  zone.c                                                      (CRC:53a86ef4)
 *      relative path: ./sys
 *      creation date: 1992
 *      last modified: Wed Dec 18 07:18:24 1996
 *      header update: Wed Dec 18 07:18:25 1996
 *  
 *       Microkernel system files
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* zone.c
 *  Zone memory allocation functions
 */

#define KERNEL
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/global.h>
#include <error.h>
#include <stdio.h>


typedef struct zone {
	struct zone *next;	/* next zone structure */
	ulong	esize;		/* size of each zone element */
	ulong	nelem;		/* total number of elements */
	ulong	nfree;		/* number of free elements */
	char	*name;		/* name of zone */
	vaddr	free;		/* start of element freelist */
	lock_t	lock;		/* is this zone being accessed? */
} Zone; 


static Zone zone_zone = {
	NULLP(struct zone), sizeof(Zone), 0, 0, "zone", (vaddr)NULL, 0
};

static Zone *zone_list = &zone_zone;




ZoneP zinit(int size, char *name)
{
Zone *zone;
	zone = (Zone*)zalloc(&zone_zone);
	zone->esize = size;
	zone->name = name;
	zone->nelem = 0;
	zone->nfree = 0;
	zone->free = (vaddr)NULL;
	zone->next = zone_list;
	zone_list = zone;
	LOCK_CLEAR(zone->lock);
	return((ZoneP)zone);
}


void zaddmem(ZoneP zone, vaddr adr, int size)
{
Zone *zp = (Zone*)zone;
int nelem;
int i;
	LOCK(zp->lock)
	nelem = size/zp->esize;
	for(i=0; i<nelem; i++) {
		*(vaddr*)adr = zp->free;
		zp->free = adr;
		adr += zp->esize;
	}
	kinfo.wasted_bytes += (size-nelem*zp->esize);
	zp->nelem += nelem;
	zp->nfree += nelem;
	UNLOCK(zp->lock) 
}

void *zalloc(ZoneP zone)
{
Zone *zp = (Zone*)zone;
int n;
vaddr free;
	if(!zp->free) {
		n = align(zp->esize, PAGE_SIZE);
		zaddmem(zp,vm_alloc(&kernel.data,n,VM_KRW),n);
	}
	LOCK(zp->lock)
	free = zp->free;
	zp->free = *(vaddr*)free;
	zp->nfree--;
	UNLOCK(zp->lock) 
	return((void*)free);
}

void zfree(ZoneP zone, void *elem)
{
Zone *zp = (Zone*)zone;
	LOCK(zp->lock)
	*(vaddr*)elem = zp->free;
	zp->free = (vaddr)elem;
	zp->nfree++;
	UNLOCK(zp->lock) 
}



void zone_status()
{
Zone *zp;
int free, total;
	zp = zone_list;
	free = total = 0;
	while(zp) {
		kprintf("%s(0x%x): esize=%d free:%d/%d used:%d/%d bytes\n",
			zp->name, (vaddr)zp, zp->esize, zp->nfree, zp->nelem,
			(zp->nelem-zp->nfree)*zp->esize,		
			zp->nelem*zp->esize);	
		free += (zp->esize*zp->nfree);
		total += (zp->esize*zp->nelem);
		zp = zp->next;
	}
	kprintf("total: %d inuse, %d free, %d wasted, out of %d bytes\n",
		total-free, free, kinfo.wasted_bytes, total);
}

