/*  osh.c                                                       (CRC:ad5bfc2c)
 *      relative path: ./usr
 *      creation date: 1992
 *      last modified: Wed May 22 21:39:48 1996
 *      header update: Wed Dec 18 07:18:30 1996
 *  
 *       User-level programs
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* osh.c
 *   object-oriented shell
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/resource.h>
#include <macro.h>
#include <object.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h> 
#include <string.h>
#include <error.h>
#include <dir.h>
#include <event.h>
#include <mdmacro.h>


typedef obj_t (*cmd_fptr)(int argc, char **argv);

#define MAXLINE	256
#define MAXARGS	32

typedef struct assoc {
	char *key;
	char *value;
	struct assoc *next;
} Assoc;

typedef struct cmd {
	char key[12];
	boolean builtin;
	cmd_fptr func;
	struct cmd *next;
} Cmd;

typedef struct job {
	int	argc;	
	char	*argv[MAXARGS];
	char	argbuf[MAXLINE];
	obj_t	pid;
	int	num;
	obj_t	in;
	obj_t	out;
	obj_t	err;
	struct job *next;
	uchar	status;
} Job;

#define JOBSTATUS_RUNNING	0
#define JOBSTATUS_STOPPED	1
#define JOBSTATUS_TERMINATED	2


Assoc *env = NULL;
Assoc *alias = NULL;
Cmd *cmd = NULL;
Job *job = NULL;
Job *curjob = NULL;
Job fg;
char **history = NULL;
ushort *histnum = NULL;
int histnext = 1;
int histpos = 0;
int histsize = 0;
boolean quit;


#define malloc(x)	((void*)_(OBJ_TASK, _mem_alloc,(x),___,___))
#define free(x)		_(OBJ_TASK, _mem_free, Ptr(x), ___,___)

#define  is_num(x)    ((x >= '0' && x <= '9') || x == '-')
#define  is_word(x)   ((x >= 'A' && x <= 'Z') || (x >= 'a' && x <= 'z') \
			|| (x == '_'))
#define  is_punc(x)   ((x) && strchr("!@#$%^*?+-/\\.:\"'",x))
#define  is_sep(x)    ((x) && strchr("&()[]<>{}|x`,=;",x))
#define  is_sym(x)    (is_num(x) || is_word(x) || is_punc(x))


extern label_t _wid, _hgt, _ctrlc, _aux;
extern label_t _dir, _size;
extern label_t _cwd, _lflag, _rnoblock, _wnoblock;
extern label_t _stdin, _stdout, _stderr;
extern label_t _pos, _msleep, _exec, _wait, _signal, _thread;

obj_t class_stream, class_file, class_buffer;


static obj_t stream_open_read(char *path)
{
obj_t fd,sd;
	fd = dir_parse(OBJ_CWD, path);
	if(err(fd))
		return(fd);
	sd = err_(fd,_read,Ptr(NULL),0,___);
	if(err(sd)) 
		sd =_(class_stream,_new,fd,___,___);
	else
		sd = fd;
	return(sd);
}

static obj_t stream_open_write(char *path)
{
obj_t fd,sd;
char buf[MAXLABEL];
	fd = dir_parse(OBJ_CWD, path);
	if(err(fd)) {
		sd = dir_parse_path(OBJ_CWD,path,buf,FALSE);
		fd = _(class_file,_new,___,___,___);
		_(sd,_create,Label(buf),fd,___);
	}
	err_(fd,_set,_size,0,___);
	sd = err_(fd,_write,Ptr(NULL),0,___);
	if(err(sd))
		sd = _(class_stream,_new,fd,___,___);
	else
		sd = fd;
	return(sd);
}

static void stream_close(obj_t fd)
{
	if(fd != OBJ_STDIN && fd != OBJ_STDOUT)
		_(fd, _free, ___, ___, ___);
}


static void combine(char *buf, int argc, char **argv)
{
int i;
	buf[0] = '\0';
	i = 0;
	while(argc--) {
		strcat(buf, argv[i]);
		strcat(buf, " ");
		i++;
	}
}


static void addbuiltin(char *key, cmd_fptr func)
{
Cmd *node;
	node = (Cmd*)malloc(sizeof(Cmd));
	strcpy(node->key, key);
	node->func = func;
	node->next = cmd;
	node->builtin = TRUE;
	cmd = node;
}

static void addcmd(char *key, cmd_fptr func)
{
Cmd *node;
	node = (Cmd*)malloc(sizeof(Cmd));
	strcpy(node->key, key);
	node->func = func;
	node->next = cmd;
	node->builtin = FALSE;
	cmd = node;
}

static cmd_fptr getcmd(char *key)
{
Cmd *p;
	p = cmd;
	while(p && strcmp(p->key, key))
		p = p->next;
	if(p)
		return(p->func);
	else
		return(NULL);	
}

static boolean builtin_cmd(char *key)
{
Cmd *p;
	p = cmd;
	while(p && strcmp(p->key, key))
		p = p->next;
	if(p)
		return(p->builtin);
	else
		return(FALSE);	
}


static void assoc_set(Assoc **list, char *key, char *value)
{
	while(*list && strcmp((*list)->key, key)) 
		list = &(*list)->next;
	if(!*list) {
		*list = (Assoc*)malloc(sizeof(Assoc));
		(*list)->next = NULL;
		(*list)->key = (char*)malloc(strlen(key)+1);
		strcpy((*list)->key, key);
	}
	(*list)->value = (char*)malloc(strlen(value)+1);
	strcpy((*list)->value, value);
}


static char *assoc_get(Assoc *list, char *key)
{
	while(list && strcmp(list->key, key))
		list = list->next;
	if(!list || !list->value)
		return("");
	else
		return(list->value);
}



static int argnext(char *wbuf, char *str, int pos)
{
boolean inquote = FALSE;
int wpos = 0;
   while(str[pos] == 32 || str[pos] == 9)
      pos++;
   while(inquote || is_sym(str[pos])) {
      if(str[pos]=='"' || str[pos]=='\'') {
         inquote = !inquote;
         pos++;
	 continue;
      }
      if(str[pos]=='\\') 
         pos++;
      wbuf[wpos++] = str[pos++];
   } 
/* XXX - might be problem if str[pos] = 0 */
   if(is_sep(str[pos]) && (!wpos))
      wbuf[wpos++] = str[pos++];
   wbuf[wpos] = '\0';
   return(pos);
}


static boolean regexp_match(char *str, char *pat)
{
	while(*str && *pat) {
		if(*pat == '?' || (*pat != '*' && *pat == *str)) {
			str++;
			pat++;
			continue;
		} 
		if(*pat != '*')
			return(FALSE);
		pat++;
		if(!*pat)
			return(TRUE);
		while(*str) {
			if(regexp_match(str,pat))
				return(TRUE);
			str++;
		}
		return(FALSE);
	}
	return(!*pat && !*str);
}


static int argparse(char *line, char **argv, char *argbuf, int trans)
{
char str[MAXLINE];
char buf[MAXLINE];
int p=0;
int n=0;
obj_t wd,id;
int index;
	while(1) {
		p=argnext(str,line,p);
		if(!str[0])
			break;
		if(strchr(str,'*') || strchr(str,'?')) {
			wd = dir_parse_path(OBJ_CWD, str, buf, FALSE);
			index =0;
			while(id=err_(wd,_getattr,index++,___,___),!err(id)) {
				ObjString(str,sizeof(str),id);
				if(str[0] == '.' && buf[0] != '.')
					continue;
				if(regexp_match(str,buf)) {
					argv[n++] = argbuf;
					strcpy(argbuf, str);
					argbuf += strlen(argbuf)+1;
				}
			}
		} else {
			argv[n++] = argbuf;
			strcpy(argbuf, str);
			argbuf += strlen(argbuf)+1;
		}
	} 
	return(n);
}


static obj_t objparse(char *str)
{
	if(str[0] == '@')
		return(Label(&str[1]));
	return(dir_parse(OBJ_ROOT, str));
}


static void call_cmd(Job *j)
{
obj_t rv;
cmd_fptr func;
obj_t obj,method,arg1,arg2,arg3;
char *c;
obj_t td,fd;
	_(OBJ_THREAD, _set,  _stdin, j->in, ___);
	_(OBJ_THREAD, _set,  _stdout, j->out, ___);
	_(OBJ_THREAD, _set,  _stderr, j->err, ___);
	if((c=strchr(j->argv[0],':'))) {
		*c = '\0';
		obj = objparse(j->argv[0]);
		method = objparse(c+1);
		arg1 = arg2 = arg3 = OBJ_NULL;
		if(j->argc > 1)
			arg1 = objparse(j->argv[1]);
		if(j->argc > 2)
			arg1 = objparse(j->argv[2]);
		if(j->argc > 3)
			arg1 = objparse(j->argv[3]);
		rv = err_(obj,method,arg1,arg2,arg3);
	} else {	
		if((func=getcmd(j->argv[0])))
			rv = func(j->argc, j->argv);
		else {
			fd = dir_parse(OBJ_CWD, j->argv[0]);
			if(err(fd)) {
				print_error(j->argv[0], fd);
				return;
			}
			td = _(OBJ_TASK, _new, ___, ___, ___);
			rv = err_(td, _exec, fd, ___, ___);
			_(td, _free, ___, ___, ___);
		}
	}
	if(err(rv)) 
		print_error(j->argv[0], rv);
}

static void job_remove(Job *j)
{
Job **p;
	if(j == &fg)
		return;
	p = &job;
	while(*p != j) 
		p = &(*p)->next;
	*p = j->next;
	free(j);
	if(curjob==j)
		curjob = job;
}


static void job_fg(Job *j)
{
int i;
char *abuf;
	fg = *j;
	abuf = fg.argbuf;
	for(i=0; i<j->argc; i++) {
		fg.argv[i] = abuf;
		strcpy(abuf, j->argv[i]);
		abuf += strlen(abuf)+1;
	}
	fg.num = 0;
}

static Job *job_bg(Job *j)
{
Job *bg;
Job *p;
int i;
char *abuf;
	if(j->num)
		bg = j;
	else {
		bg = (Job*)malloc(sizeof(Job));
		*bg = *j;
		abuf = bg->argbuf;
		for(i=0; i<j->argc; i++) {
			bg->argv[i] = abuf;
			strcpy(abuf, j->argv[i]);
			abuf += strlen(abuf)+1;
		}
		bg->num = 1;
		p = job;
		while(p) {
			if(p->num == bg->num) {
				bg->num++;
				p = job;
			} else
				p = p->next;
		}
		bg->next = job;
		job = bg;
	}
	curjob = bg;
	return(bg);
}


/*
static void job_kill(Job *j)
{
	if(!j->num) 
		return;
}
*/


static char *history_match(char *str)
{
int n;
	n = histpos - 1;
	if(n < 0)
		n += histsize;	
	if(n < 0)
		return("");
	if(strcmp(str,"!")) {
		while(n != histpos) {
			if(history[n] && !strncmp(history[n],str,strlen(str)))
				break;
			if(histnum[n] == atoi(str))
				break;
			if(--n < 0)
				n = histsize - 1;
		}
	}
	if(n != histpos && history[n])
		return(history[n]);
	else
		return("");
}

static void history_update(char *str)
{
	if(!histsize)
		return;
	histnum[histpos] = histnext++;
	if(history[histpos])
		free(history[histpos]);
	history[histpos] = (char*)malloc(strlen(str)+1);
	strcpy(history[histpos],str);
	if(++histpos >= histsize)
		histpos = 0;
}

static void job_show(Job *j)
{
int i;
	if(j->num) {
		printf("[%d] %x ", j->num, j->pid);
		switch(j->status) {
		case JOBSTATUS_RUNNING:
			printf("Running\t"); break;
		case JOBSTATUS_STOPPED:
			printf("Stopped\t"); break;
		case JOBSTATUS_TERMINATED:
			printf("Terminated\t"); break;
		}
	}
	for(i=0; i<j->argc; i++)
		printf("%s ",j->argv[i]);
	printf("\n");
}

static Job *job_find(obj_t pid)
{
Job *j;
	j = job;
	while(j && j->pid != pid)
		j = j->next;
	return(j);	
}


static void wait_for_termination(obj_t pid)
{
Event event;
Job *bg;
	_(OBJ_STDERR,_set,_thread,pid,___);
	while(1) {
		_(OBJ_THREAD,_wait,EVENT_ANY,PARAM_ANY,Ptr(&event));
		if(event.event==EVENT_CHILDTERM) {
			if(event.sender == pid)
				break;
			bg = job_find(event.sender);
			bg->status = JOBSTATUS_TERMINATED;
			job_show(bg);
			job_remove(bg);
		} else if(event.event==EVENT_CHILDSTOP) {
			if(event.sender == pid) {
				bg = job_bg(&fg);
				bg->status = JOBSTATUS_STOPPED;	
				job_show(bg);
				break;
			} else
				printf("Unexpected EVENT_CHILDSTOP:%x\n",
					event.sender);
		} else 
			printf("Unexpected event: %x from %x\n", 
					event.event,event.sender);
	}
	_(OBJ_STDERR,_set,_thread,OBJ_NULL,___);
}


static void parse_line(char *cmdline)
{
char buf[MAXLINE];
Job *bg;
int argc;
char *argv[MAXARGS];
int n,i,nargs;
char *astr;
char *abuf;
char *av[MAXARGS];
obj_t def_stdin, def_stdout, def_stderr;
char *cp,*dp,*tp;
char tmp[MAXLABEL];
	if(!*cmdline || cmdline[0] == '#')
		return;
	cp = cmdline;
	dp = buf;
	while(*cp) {
		if(*cp == '\\') {
			cp++;
			*dp++ = *cp++;
		} else if(*cp == '$') {	/* environment variable substitution  */
			cp++;
			tp = tmp;
			while(*cp >= 32)
				*tp++ = *cp++;
			*tp = '\0';
			strcpy(dp,assoc_get(env,tmp));
			while(*dp)
				dp++;
		} else if(*cp == '!') {	/* history substitution */
			cp++;
			tp = tmp;
			while(*cp >= 32)
				*tp++ = *cp++;
			*tp = '\0';
			strcpy(dp,history_match(tmp));
			while(*dp)
				dp++;
		} else	
			*dp++ = *cp++;
	}
	*dp = '\0';

	abuf = fg.argbuf;
	argc = argparse(buf, argv, fg.argbuf, 1);
	abuf += strlen(buf)+argc;
	if(!argc)
		return;

	history_update(buf);

	n = 0;
	fg.argc = 0;
	fg.num = 0;
	def_stdin = _(OBJ_THREAD, _get, _stdin, ___, ___);
	def_stdout = _(OBJ_THREAD, _get, _stdout, ___, ___);
	def_stderr = _(OBJ_THREAD, _get, _stderr, ___, ___);
	fg.in = def_stdin;
	fg.out = def_stdout;
	fg.err = def_stderr;
	fg.status = JOBSTATUS_RUNNING;
	while(n < argc) {
		if(!strcmp(argv[n],"&")) {
			bg = job_bg(&fg);
			bg->pid=_(OBJ_THREAD,_new,Label(bg->argv[0]),
				Ptr(call_cmd),Ptr(bg));
			job_show(bg);
			fg.in = def_stdin;
			fg.out = def_stdout;
			fg.err = def_stderr;
			fg.argc = 0;
		} else if(!strcmp(argv[n],";")) {
			if(builtin_cmd(fg.argv[0]))
				call_cmd(&fg);
			else {
				fg.pid=_(OBJ_THREAD,_new,Label(fg.argv[0]),
					Ptr(call_cmd),Ptr(&fg));
				wait_for_termination(fg.pid);
			}
			fg.in = def_stdin;
			fg.out = def_stdout;
			fg.err = def_stderr;
			fg.argc = 0;
		} else if(!strcmp(argv[n],">")) {
			n++;
			fg.out = stream_open_write(argv[n]);
		} else if(!strcmp(argv[n],"<")) {
			n++;
			fg.in = stream_open_read(argv[n]);
		} else if(!strcmp(argv[n],"|")) {
			fg.out = _(class_buffer, _new, ___, ___, ___);
			bg = job_bg(&fg);
			bg->pid=_(OBJ_THREAD,_new,Label(bg->argv[0]),
					Ptr(call_cmd),Ptr(bg));
			fg.in = fg.out;
			fg.out = def_stdout;
			fg.err = def_stderr;
			fg.argc = 0;
		} else {
			if(!fg.argc) {	/* check aliases: */
				while((astr=assoc_get(alias,argv[n])) && astr[0]) {
					nargs = argparse(astr,av,abuf,1);
					for(i=argc-1; i>n; i--)
						argv[i+nargs-1] = argv[i];
					for(i=0; i<nargs; i++)
						argv[n+i] = av[i];
					argc += nargs-1;
					abuf += strlen(astr)+nargs;
				}
			}
			fg.argv[fg.argc++] = argv[n];
		}
		n++;	
	}
	if(fg.argc > 0) {
		if(builtin_cmd(fg.argv[0]))
			call_cmd(&fg);
		else {
			fg.pid = _(OBJ_THREAD,_new,Label(fg.argv[0]),
				Ptr(call_cmd),Ptr(&fg));
			wait_for_termination(fg.pid);
		}
	}
}


static void prompt_print(void)
{
char *cp,*dp;
char buf[MAXLINE];	
	cp = assoc_get(env,"prompt");
	dp = buf;	
	while(*cp) {
		if(*cp == '!') {
			sprintf(dp,"%d",histnext);
			dp += strlen(dp);
		} else
			*dp++ = *cp;
		cp++;
	}	
	*dp = '\0';
	printf("%s",buf);
}


static Job *get_job(char *s)
{
int n;
Job *j;
	if(s[0]=='%')
		s++;
	n = atoi(s);
	j = job;
	while(j && j->num != n)
		j = j->next;
	return(j);
}

static obj_t get_pid(char *s)
{
int n;
Job *j;
obj_t wd;
	if(s[0] == '%') {
		n = atoi(&s[1]);
		j = job;
		while(j && j->num != n)
			j = j->next;
		if(!j)	
			return(OBJ_NULL); 
		else
			return(j->pid);
	}
	wd = _(OBJ_ROOT, _get, Label("thread"), ___, ___);
	return(err_(wd, _get, atoi(s), ___, ___));
}


static obj_t cmd_kill(int argc, char **argv)
{
obj_t td;
int i;
	for(i=1; i<argc; i++) {
		td = get_pid(argv[i]);
		_(td, _free, ___, ___, ___);
	}
	return(OK);
}


static obj_t cmd_set(int argc, char **argv)
{
Assoc *p;
int i,n;
	if(argc == 1) {
		p = env;
		while(p) {
			printf("%s\t%s\n", p->key, p->value);
			p = p->next;
		}
		return(OK);
	}
	i = 0;
	while(++i < argc) {
		if(i == argc-1 || strcmp(argv[i+1],"=")) 
			assoc_set(&env, argv[i], "");
		else {
			assoc_set(&env, argv[i], argv[i+2]);
			if(!strcmp(argv[i],"history")) {
				for(n=0; n<histsize; n++)
					free(history[n]);
				if(history) {
					free(history);
					free(histnum);
				}
				histsize = atoi(argv[i+2]);
				history = (char**)malloc(histsize*sizeof(char*));
				histnum = (ushort*)malloc(histsize*sizeof(ushort));
				for(n=0; n<histsize; n++)
					history[n] = NULL;
				histpos = 0;
			}
			i += 2;
		}
	}
	return(OK);
}


static obj_t cmd_setenv(int argc, char **argv)
{
	if(argc == 3)
		assoc_set(&env, argv[1], argv[2]);
	return(OK);
}

static obj_t cmd_unset(int argc, char **argv)
{
	return(OK);
}


static obj_t cmd_alias(int argc, char **argv)
{
Assoc *p;
char buf[MAXLINE];
	if(argc == 1) {
		p = alias;
		while(p) {
			printf("%s\t%s\n", p->key, p->value);
			p = p->next;
		}
		return(OK);
	}
	if(argc == 2) {
		printf("%s\n",assoc_get(alias,argv[1]));
		return(OK);
	}
	combine(buf,argc-2,&argv[2]);
	assoc_set(&alias, argv[1], buf);
	return(OK);
}

static obj_t cmd_unalias(int argc, char **argv)
{
	return(OK);
}


static obj_t cmd_echo(int argc, char **argv)
{
int n;
boolean nl = TRUE;
	n = 1;
	if(!strcmp(argv[n],"-n")) {
		nl = FALSE;
		n++;
	}
	while(n<argc) {
		printf("%s ", argv[n]);
		n++;
	}
	if(nl)
		printf("\n");
	return(OK);
}


static obj_t cmd_jobs(int argc, char **argv)
{
Job *j;
	j = job;
	while(j) {
		job_show(j);
		j = j->next;
	}
	return(OK);
}

static obj_t cmd_fg(int argc, char **argv)
{
obj_t pid;
Job *bg;
	pid = _(OBJ_THREAD,_eval,___,___,___);
	if(argc == 1)
		bg = curjob;
	else
		bg = get_job(argv[1]);
	if(!bg) {
		eprintf("fg: No such job\n");
		return(OK);
	}	
	job_fg(bg);
	job_remove(bg);
	job_show(&fg);
	_(fg.pid,_signal,EVENT_WAKEUP,0,pid);
	wait_for_termination(fg.pid);
	return(OK);
}

static obj_t cmd_bg(int argc, char **argv)
{
obj_t pid;
Job *bg;
int i;
	pid = _(OBJ_THREAD,_eval,___,___,___);
	for(i=0; i<argc; i++) {
		if(!i)
			if(argc==1)
				bg = curjob;
			else
				continue;
		else	
			bg = get_job(argv[i]);
		if(!bg) {
			eprintf("bg: No such job\n");
			return(OK);
		}	
		bg->status = JOBSTATUS_RUNNING;
		_(bg->pid,_signal,EVENT_WAKEUP,0,pid);
		job_show(bg);
	}
	return(OK);
}

static obj_t cmd_source(int argc, char **argv)
{
obj_t fd;
char buf[MAXLINE];
	fd = stream_open_read(argv[1]);
	if(err(fd))
		return(fd);
	while(ogets(buf,sizeof(buf),fd)) 
		parse_line(buf);
	stream_close(fd);
	return(OK);
}


static obj_t cmd_history(int argc, char **argv)
{
int i,n;
	i = histsize;
	n = histpos;
	while(i--) {
		if(history[n])
			printf("%4d  %s\n", histnum[n], history[n]);
		if(++n >= histsize)
			n = 0;
	}
	return(OK);
}


static obj_t cmd_exit(int argc, char **argv)
{
	quit = TRUE;
	return(OK);
}


static obj_t cmd_cd(int argc, char **argv)
{
obj_t wd;
	wd = dir_parse(OBJ_CWD,argv[1]);
	if(err(wd))	
		return(wd);
	_(OBJ_THREAD,_set,_cwd,wd,___);
	assoc_set(&env,"cwd",argv[1]);
	return(OK);
}



/*  higher-level commands--- this should probably be separate programs, 
 *  but for now, they're imbedded in the shell:
 */

static obj_t cmd_cat(int argc, char **argv)
{
int n, i;
char buf[2048];
obj_t fd;
	for(i=0; i<argc; i++) {
		if(!i)
			if(argc == 1)
				fd = OBJ_STDIN;
			else
				continue;
		else {
			fd = stream_open_read(argv[i]);
			if(err(fd)) 
				return(fd);
		}
		while((n=err_(fd,_read,Ptr(buf),sizeof(buf),___))) {
			if(err(n)) {
				print_error(argv[i], n);
				break;
			}
			_(OBJ_STDOUT,_write,Ptr(buf),n,___);
		}
		stream_close(fd);
	}
	return(OK);
}


static obj_t cmd_sleep(int argc, char **argv)
{
	_(OBJ_THREAD, _msleep, 1000*atoi(argv[1]), ___, ___);
	return(OK);
}

static obj_t cmd_in(int argc, char **argv)
{
#ifdef I386
	printf("in(%x)=%x\n", atoi(argv[1]), in_portb(atoi(argv[1])));
#endif
	return(OK);
}

static obj_t cmd_out(int argc, char **argv)
{
#ifdef I386
	out_portb(atoi(argv[1]), atoi(argv[2]));
	printf("out %x,%x\n", atoi(argv[1]), atoi(argv[2]));
#endif
	return(OK);
}


static obj_t cmd_loop(int argc, char **argv)
{
	while(1)
		printf("hello world...");
	return(OK);
}



static obj_t cmd_hog(int argc, char **argv)
{
int x;
	while(1)
		x = x*x;
	return(OK);
}


static obj_t cmd_x(int argc, char **argv)
{
vaddr adr;
ulong val;
   adr = strtoul(argv[1],0,0);
   if(argc == 2)
   	printf("*0x%x = %x\n", adr, *(ulong*)adr);
   else {
        val = strtoul(argv[2],0,0);
   	printf("*0x%x = %x\n", adr, val);
	*(ulong*)adr = val;
   }
   return(OK);	
}


static obj_t cmd_clear(int argc, char **argv)
{
	printf("\013\014");
	 /* puts("\033[2J\033[H"); */
	return(OK);
}


static obj_t cmd_mkdir(int argc, char **argv)
{
obj_t wd,nd;
char buf[MAXLABEL];
int i;
	for(i=1; i<argc; i++) {
		wd = dir_parse_path(OBJ_CWD,argv[i],buf,FALSE);
		if(err(wd)) {
			print_error(argv[1], wd);
			continue;
		}
		nd = _(wd,_new,___,___,___);
		_(wd,_create,Label(buf),nd,___);
	}
	return(OK);
}

		

static obj_t cmd_mv(int argc, char **argv)
{
obj_t src,wd;
char buf[MAXLABEL];
	src = dir_parse(OBJ_CWD,argv[1]);
	if(err(src)) {
		print_error(argv[1], src);
		return(OK);
	}
	wd = dir_parse_path(OBJ_CWD,argv[2],buf,FALSE);
	if(err(wd)) {
		print_error(argv[2], wd);
		return(OK);
	}
	_(wd,_create,Label(buf),src,___);
	wd = dir_parse_path(OBJ_CWD,argv[1],buf,FALSE);
	wd = _(wd,_remove,Label(buf),___,___);
	if(err(wd))
		printf("cannot remove %s: %s\n", buf, strerror(wd));
	return(OK);
}



static obj_t cmd_ln(int argc, char **argv)
{
obj_t src,wd;
char buf[MAXLABEL];
	src = dir_parse(OBJ_CWD,argv[1]);
	if(err(src)) {
		print_error(argv[1], src);
		return(OK);
	}
	wd = dir_parse_path(OBJ_CWD,argv[2],buf,FALSE);
	if(err(wd)) {
		print_error(argv[2], wd);
		return(OK);
	}
	_(wd,_create,Label(buf),src,___);
	return(OK);
}


static obj_t cmd_cp(int argc, char **argv)
{
obj_t src,wd,new;
char buf[MAXLABEL];
	src = dir_parse(OBJ_CWD,argv[1]);
	if(err(src)) {
		print_error(argv[1], src);
		return(OK);
	}
	wd = dir_parse_path(OBJ_CWD,argv[2],buf,FALSE);
	if(err(wd)) {
		print_error(argv[2], wd);
		return(OK);
	}
	new = _(src,_dup,___,___,___);
	_(wd,_create,Label(buf),new,___);
	return(OK);
}


static void ls_formatted(obj_t wd, char *sw, obj_t attr[],
				char fmt[][MAXLABEL], int n)
{
obj_t fd, id, val;
int index = 0;
char buf[MAXLABEL];
char tmp[MAXLABEL];
int i;
char *cp,*sp;
char *line[600];	/* XXX - should be fully dynamic */
char curline[MAXLINE];
int lnum;
int len, maxlen;
int wid;
	lnum = 0;
	maxlen = 0;
	wid = _(OBJ_STDOUT, _get, _wid, ___, ___);
	while(1) {
		cp = curline;
		id = err_(wd,_getattr, index++, ___, ___);
		if(id == ERR_EOA)
			break;
		ObjString(buf,sizeof(buf),id);
		if(!strchr(sw,'a') && buf[0]=='.') 
			continue;
		fd = err_(wd,_get,id,___,___);
		if(err(fd)) {
			print_error(buf, fd);
			break;
		}
		for(i=0; i<n; i++) {
			if(attr[i] == OBJ_NULL)
				sprintf(cp,fmt[i],buf);
			else if(attr[i] == OBJ_CWD) {
				if(strchr(fmt[i],'s')) {
					ObjString(tmp,sizeof(tmp),fd);
					sprintf(cp,fmt[i],tmp);
				} else
					sprintf(cp,fmt[i],fd);
			} else {
				val = err_(fd,_get,attr[i],___,___); 
				if(strchr(fmt[i],'s')) {
					ObjString(tmp,sizeof(tmp),val);
					sprintf(cp,fmt[i],tmp);
				} else
					sprintf(cp,fmt[i],val);
			}
			while(*cp)
				cp++;
		}
		len = strlen(curline);
		if(len > maxlen)	
			maxlen = len;
		line[lnum] = malloc(len);
		strcpy(line[lnum++], curline);
	}
	for(i=0; i<lnum-1; i++) for(n=i+1; n<lnum; n++)
		if(strcmp(line[i],line[n]) > 0) {
			cp = line[i];
			line[i] = line[n];
			line[n] = cp;
		}
	if(strchr(sw,'C')) {
		maxlen = align(maxlen+2, 4);
		i = wid/maxlen;
		n = (lnum-1)/i+1;
		wid = maxlen*(lnum/n);
	} else
		maxlen = wid;
	len = 0;
	for(i=0; i<lnum; i++) {
		cp = curline;
		sp = line[i];
		n = maxlen;
		len += maxlen;
		while(*sp) {
			*cp++ = *sp++;
			n--;
		}
		while(n > 0 && len < wid)  {
			*cp++ = ' ';
			n--;
		}
		*cp = '\0';
		if(len >= wid) {
			len = 0;
			printf("%s\n", curline);
		} else
			_(OBJ_STDOUT, _write, Ptr(curline), maxlen+n, ___);	
		free(line[i]);
	}
	if(len)
		printf("\n");
}


/* XXX - should take a command script that lists a different
 *  format for each filetype (dir/etc)
 */
static obj_t cmd_lf(int argc, char **argv)
{
obj_t attr[20];
char fmt[20][MAXLABEL];
int i,n;
char *cp,*dp;
char buf[MAXLABEL];
char sw[40];
obj_t wd;
	i = 1;
	n = 0;
	sw[0] = '\0';
	while(i < argc) {
		if((cp=strchr(argv[i],':'))) {
			if(cp==argv[i])
				attr[n] = OBJ_NULL;
			else {
				strncpy(buf,argv[i],cp-argv[i]);
				if(!strcmp(buf,"_"))
					attr[n] = OBJ_CWD;
				else
					attr[n] = Label(buf);
			}
			cp++;
			dp = fmt[n];
			while(*cp) {
				if(*cp == '_')
					*dp++ = ' ';
				else
					*dp++ = *cp;
				cp++;
			}
			*dp = '\0';
			n++;
		} else if(argv[i][0] == '-') 
			strcat(sw,&argv[i][1]);
		else
			break;
		i++;
	}
	if(!n) {
		strcpy(fmt[0],"%s");
		attr[0] = OBJ_NULL;
		n = 1;
	}
	if(i==argc) 
		ls_formatted(OBJ_CWD, sw, attr, fmt, n);
	if(i==argc-1) {
		wd = dir_parse(OBJ_CWD, argv[i++]);
		ls_formatted(wd,sw,attr,fmt,n);
	}
	while(i < argc) {
		wd = dir_parse(OBJ_CWD,argv[i]);
		if(err_(wd,_type,_dir,___,___) == OBJ_TRUE)
			printf("%s:\n",argv[i]);
		ls_formatted(wd,sw,attr,fmt,n);
		if(++i < argc)
			printf("\n");
	}
	return(OK);
}


static obj_t cmd_rm(int argc, char **argv)
{
int i = 0;
obj_t rv;
obj_t wd;
char buf[MAXLABEL];
	while(++i < argc) {
		wd = dir_parse_path(OBJ_CWD,argv[i],buf,FALSE);
		if(err(wd)) {
			print_error(argv[1], wd);
			continue;
		}
		rv = err_(wd,_remove,Label(buf),___,___);
		if(err(rv))
			print_error(buf, rv);
	}
	return(OK);
}


static obj_t cmd_dd(int argc, char **argv)
{
int pos;
int size;
char buf[4096];
obj_t fd;
int n;
	if(argc == 4) {	
		pos = atoi(argv[1]);
		size = atoi(argv[2]);
		fd = dir_parse(OBJ_CWD, argv[3]);
	} else {
		pos = atoi(argv[1]);
		size = atoi(argv[2]);
		fd = OBJ_STDIN;
	}
	size += pos;
	cprintf("reading from %d to %d\n", pos, size);
	while((n = err_(fd,_read,Ptr(buf),sizeof(buf),pos))) {
		cprintf("read %d,%d bytes\n", n, min(n,size-pos));
		_(OBJ_STDOUT,_write,Ptr(buf),min(n,size-pos),pos);
		pos += n;
		if(pos >= size)
			break;
	}
	return(OK);
}


static obj_t cmd_od(int argc, char **argv)
{
obj_t fd;
int x, n, pos, i;
uchar buf[4096];
	if(argc < 2)
		fd = OBJ_STDIN;
	else
		fd = stream_open_read(argv[1]);
	x = 0;
	pos = 0;
	while((n = err_(fd,_read,Ptr(buf),sizeof(buf),pos))) {
		if(err(n)) {
			print_error(argv[1], n);
			return(OK);
		}
		for(i=0; i<n; i++) {
			if(x == 0)
				printf("%8x: ",pos+i);
			printf("%2x ", buf[i]);
			buf[i] &= 0x7f;
			if(buf[i] < 32)
				buf[i] = '.';
			if(++x >= 16) {
				x = 0;
				printf("%16s\n",&buf[i-15]);
			}
		}
		pos += n;
  	} 
	if(x) {
		while(++x < 16) {
			printf("   ");
			buf[n] = ' ';
			n++;
		} 
		printf("%16s\n",&buf[n-15]);
	}
	return(OK);
}


static obj_t cmd_more(int argc, char **argv)
{
int n, i;
char ch, temp;
obj_t fd;
int wid,hgt,x,y;
int lflag;
obj_t tty;
	tty = _(OBJ_THREAD, _get, _stderr,___,___);	/* XXX - should be tty */
  	wid = _(tty, _get, _wid,___,___);
	hgt = _(tty, _get, _hgt,___,___);
	x = y = 0;
	for(i=0; i<argc; i++) {
		if(!i) 
			if(argc == 1)
				fd = OBJ_STDIN;
			else
				continue;
		else {
			fd = stream_open_read(argv[i]);
			if(err(fd)) {
				print_error(argv[i], fd);
				return(OK);
			}
		}
		while((n = err_(fd,_read,Ptr(&ch),1,___))) {
			if(err(n)) {
				print_error(argv[i], n);
				return(OK);
			}
			if (ch == '\n' || x++ > wid ) { 
				y++;
				x = 0;
				if (y >= hgt) {  
					lflag = _(tty, _get, _lflag,___,___);
					_(tty, _set, _lflag, 0,___);
					while(_(tty, _read, Ptr(&temp),
							 1, ___) < 1)
						_(OBJ_THREAD, _msleep, 0, ___, ___);
					_(tty, _set, _lflag, lflag, ___);
					y = 0;
				}
			}
			_(OBJ_STDOUT,_write,Ptr(&ch),1,___);
		}
		if(fd != OBJ_STDIN)
			stream_close(fd);  
  	} 
	return(OK);
}



static obj_t cmd_who(int argc, char **argv)
{
   printf("load average: ");
   _(OBJ_SYS, Label("load_average"), ___, ___, ___);
   puts("User\ttty\tWhat");
   puts("root\t00\twho");
   return(OK);
}




static void init_cmd(void)
{
	addbuiltin("set",cmd_set);
	addbuiltin("unset",cmd_unset);
	addbuiltin("alias",cmd_alias);
	addbuiltin("unalias",cmd_unalias);
	addbuiltin("printenv",cmd_set);
	addbuiltin("setenv",cmd_setenv);
	addbuiltin("logout",cmd_exit);
	addbuiltin("exit",cmd_exit);
	addbuiltin("source",cmd_source);
	addbuiltin("kill",cmd_kill);
	addbuiltin("jobs",cmd_jobs);
	addbuiltin("fg",cmd_fg);
	addbuiltin("bg",cmd_bg);
	addbuiltin("cd",cmd_cd);
	addbuiltin("history",cmd_history);

	addcmd("echo",cmd_echo);
	addcmd("sleep",cmd_sleep);
	addcmd("cat",cmd_cat);
	addcmd("lf",cmd_lf);
	addcmd("mkdir",cmd_mkdir);
	addcmd("clear",cmd_clear);
	addcmd("ln",cmd_ln);
	addcmd("rm",cmd_rm);
	addcmd("cp",cmd_cp);
	addcmd("mv",cmd_mv);
	addcmd("who",cmd_who);
	addcmd("w",cmd_who);
	addcmd("more",cmd_more);
	addcmd("lf",cmd_lf);
	addcmd("od",cmd_od);
	addcmd("dd",cmd_dd);
	
	addcmd("hog",cmd_hog);
	addcmd("x",cmd_x);
	addcmd("loop",cmd_loop);
	addcmd("in",cmd_in);
	addcmd("out",cmd_out);
}


obj_t Main(ObjFrame *frame, vaddr base, obj_t argdir, obj_t envdir)
{
char buf[MAXLINE];
char *argv[2];
	class_file = dir_parse(OBJ_ROOT, "/class/File");
	class_stream = dir_parse(OBJ_ROOT, "/class/Stream");
	class_buffer = dir_parse(OBJ_ROOT, "/class/Buffer");
	init_cmd();
	assoc_set(&env,"user","root");
	assoc_set(&env,"status","0");
	assoc_set(&env,"argv","()");
	assoc_set(&env,"home","/");
	assoc_set(&env,"cwd","/");
	assoc_set(&env,"host","fugazi");
	assoc_set(&env,"shell","osh");
	assoc_set(&env,"prompt","% ");
	quit = FALSE;
	argv[0] = "source";
	argv[1] = "/.oshrc";
	cmd_source(2,argv);
	while(1) {
		if(quit) 
			return(OK);
		prompt_print();	
		if(!getstrn(buf,sizeof(buf))) {
			puts("exit\n");
			return(OK);
		}
		parse_line(buf);
	}
	return(OK);
}

